package edu.wpi.cs.dsrg.xmldb.xat.common.operator;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
/**
 * Implementation of the XATQueryObject interface.  Any object that
 * wants to have a reference to the statistics gatherer -or- be able to be
 * stored in the StatisticsGatherer needs to derive from this class. <p>
 *
 * This also provides a standard means for operators to define which statistics they support.
 * This is new b/c previously, the properties and statistics were lumped together.  
 *
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0        
 */
public abstract class XATQueryObjectImp
    extends XATPropertiesImp
    implements XATQueryObject, Comparable {

    protected boolean statsPresent;

    /** each instance of this class also has the StatisticsGatherer available for use*/
    protected edu
        .wpi
        .cs
        .dsrg
        .xmldb
        .xat
        .component
        .statisticsgatherer
        .StatisticsGatherer stats;
    /** this actually stores the statistics supported */
    private java.util.Properties statistics;

    /**
     * Instantiates the XATQueryObject
     * @since   1.0            
     */
    public XATQueryObjectImp() {
        super();
        statistics = new java.util.Properties();
        setDefaultStatistics();
        String prop = System.getProperty("StatisticsGatherer");
        if (prop == null) {
            statsPresent = false;
        } else {
            statsPresent = true;
        }

        if (statsPresent)
            stats =
                edu
                    .wpi
                    .cs
                    .dsrg
                    .xmldb
                    .xat
                    .component
                    .statisticsgatherer
                    .StatisticsGatherer
                    .getStatisticsGather();
    }
    /**
    * Adds a new Statistic that will be supported  
    * @param name java.lang.String the name of the property
    * @since   1.0            
    */
    protected void addStatistic(String name) {
        statistics.setProperty(name, "");

    }
    /**
    * Note: this class has a natural ordering that is inconsistent with equals.
    * This is sorted based on hashcode only, it does not have anything to do with the contents
    * of the query object.  Subclasses should override this to have a more correct compareTo. <p>
    *
    * Compares this object with the specified object for order.  Returns a
    * negative integer, zero, or a positive integer as this object is less
    * than, equal to, or greater than the specified object.<p>
    *
    * The implementor must ensure <tt>sgn(x.compareTo(y)) ==
    * -sgn(y.compareTo(x))</tt> for all <tt>x</tt> and <tt>y</tt>.  (This
    * implies that <tt>x.compareTo(y)</tt> must throw an exception iff
    * <tt>y.compareTo(x)</tt> throws an exception.)<p>
    *
    * The implementor must also ensure that the relation is transitive:
    * <tt>(x.compareTo(y)&gt;0 &amp;&amp; y.compareTo(z)&gt;0)</tt> implies
    * <tt>x.compareTo(z)&gt;0</tt>.<p>
    *
    * Finally, the implementer must ensure that <tt>x.compareTo(y)==0</tt>
    * implies that <tt>sgn(x.compareTo(z)) == sgn(y.compareTo(z))</tt>, for
    * all <tt>z</tt>.<p>
    *
    * It is strongly recommended, but <i>not</i> strictly required that
    * <tt>(x.compareTo(y)==0) == (x.equals(y))</tt>.  Generally speaking, any
    * class that implements the <tt>Comparable</tt> interface and violates
    * this condition should clearly indicate this fact.  The recommended
    * language is "Note: this class has a natural ordering that is
    * inconsistent with equals."
    * 
    * @param   o the Object to be compared.
    * @return  a negative integer, zero, or a positive integer as this object
    *		is less than, equal to, or greater than the specified object.
    * 
    * @throws ClassCastException if the specified object's type prevents it
    *         from being compared to this Object.
    * @since   1.0           
    */
    public int compareTo(java.lang.Object o) {
        Long thisCode = new Long(this.hashCode());
        Long otherCode = new Long(o.hashCode());
        return thisCode.compareTo(otherCode);
    }
    /**
    * returns an enumeration of Statistics supported by
    * this query object
    * @return an Enumeration
    * @since   1.0           
    */
    public java.util.Enumeration getStatistics() {
        return statistics.keys();
    }
    /**
    * Checks if the given property is valid for this QueryObject.  
    * @return boolean true if the property is supported by this object, false otherwise.
    * @since   1.0            
    */
    public boolean isValidStatistic(String nameToCheck) {
        return statistics.containsKey(nameToCheck);
    }
    /**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    * @since   1.0           
    */
    protected void setDefaultProperties() {
        //the weight of the previous values in calculating average
        addProperty("AVERAGE_WEIGHT", ".7");
    }
    /**
    * called by the constructor.  this will set the default Statistics
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultStatistics() so that the common statistics may be set.
    * @since   1.0           
    */
    protected void setDefaultStatistics() {
    }
}
