package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.Set;

import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;

/**
 * The class defines the node of XATTree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public interface XATNode extends Visitable {
    /**
     * Add one child to this node. This should be called only by the XATTree.addNode().
     * Remember the children are order unsensitive.
     * @return int The number of children.
     * @param new_child The new child node added.
     * @since 1.0
     */
    int addChild(XATNode new_child);
    /**
     * Add one child to this node at specified position. 
     * This method should be called only by the XATTree.addNode().
     * The children are order unsensitive.
     * @return The number of children.
     * @param new_child The new child node.
     * @param position The position of new added node.
     * @since 1.0
     */
    int addChild(XATNode new_child, int position);
    /**
     * Clear the schema of this XATNode.
     * @return Set The old schema of this XATNode.
     * @since 1.0
     */
    Set clearSchema();
    /**
     * Deep clone this XATNode.
     * @return Object The cloned result.
     * @since 1.0
     */
    public Object clone();
    /**
     * Delete a child from this node.
     * @return int The number of children.
     * @param child The to-be-deleted child node.
     * @since 1.0
     */
    int deleteChild(XATNode child);
    /**
     * Get a specific child node of this node by their position.
     * @return XATNode The child node got.
     * @param position The position of the child node.
     * @throws XATNodeDoesNotExistException
     * @since 1.0
     */
    XATNode getChild(int position) throws XATNodeDoesNotExistException;
    /** 
     * Get all the children nodes of this node.
     * @return XATNode[] The array of all the children.
     * @since 1.0
     */
    XATNode[] getChildren();
    /**
     * Returns how many children this node has
     * @return int The number of children.
     * @since 1.0
     */
    int getChildrenSize();
    /**
     * Return the index of a node.
     * @return int The index number of given node.
     * @param node The XATNode computing index for.
     * @since 1.0
     */
    int getIndexOf(XATNode node);
    /** 
     * Get the operator in this Node.
     * @return XATOperator The operator of this node.
     * @since 1.0
     */

    XATOperator getOperator();
    /** 
     * Get the parent node of this node.
     * @return XATNode The parent of this node.
     * @since 1.0
     */

    XATNode getParent();
    /**
     * Get an Iterator of all of the parents.  
     * The XATNode only has one parent, thus this iterator will only contain one element. 
     * However, subclasses may have multiple parents, thus nothing is lost by placing this method here.
     * @return Iterator Including all the parents.
     * @since 1.0
     */
    java.util.Iterator getParents();
    /**
     * Get the schema of this XATNode.
     * @return Set The schema of this XATNode.
     * @since 1.0
     */

    Set getSchema();
    /**
     * Get a position for a specific child node.
     * If not found, return -1.
     * @return int The position of this child.
     * @param child The reference of child node searching for.
     * @since 1.0
     */
    int locateChild(XATNode child);
    /**
     * Set a specific child to be a new child node.
     * @return XATNode The original child node.
     * @param position The position of the child node.
     * @param new_child The new child node.
     * @since 1.0
     */
    XATNode setChild(int position, XATNode new_child)
        throws XATNodeDoesNotExistException;
    /**
     * Set all the children nodes of this node.
     * @return XATNode[] The array containing all the old children.
     * @param new_children The new children nodes.
     * @since 1.0
     */
    XATNode[] setChildren(XATNode[] new_children);
    /**
     * Set the operator in this Node.
     * @return XATOperator The old operator in this node.
     * @param new_operator The new operator in this node.
     * @since 1.0
     */
    XATOperator setOperator(XATOperator new_operator);
    /**
     * Set the parent node of this node.
     * @return XATNode The original parent of this node.
     * @param new_parent The new parent node of this node.
     * @since 1.0
     */
    XATNode setParent(XATNode new_parent);
    /**
     * Set the schema for current node.
     * @return Set The old schema of this node.
     * @param new_schema The new schema for this node.
     * @since 1.0
     */
    Set setSchema(Set new_schema);
}
