package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;

import edu.wpi.cs.dsrg.xmldb.xat.common.expression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.*;

/**
 * This Class is used to push down operators by traversing the XAT.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

 public class ComputationPushdownRules extends ReWriteRules {
/**
 * Default ComputationPushdownRules constructor.
 * Constructs a ComputationPushdownRules object.
 * @since 1.0   
 */

public ComputationPushdownRules() {
    super();
}
/**
 * Clean all the schema of XAT with a given root.
 * This method recursively clean the schema of XAT node. Set the schema of each node as null.
 * @return void 
 * @param root edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 * @since 1.0   
 * @see   edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @see   edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
protected void cleanSchema(XATNode root) {
    root.clearSchema();
    // set all the children.
    XATNode[] children = root.getChildren();
    for (int i = 0; i < children.length; i++) {
        cleanSchema(children[i]);
    }

    // set all the subtree.
    if (root instanceof FOR) {
        XATTree subtree = ((FOR) root).getXtreFunctions();
        cleanSchema(subtree.getRoot());
    } else
        if (root instanceof GroupBy) {
            XATTree subtree = ((GroupBy) root).getXATTree();
            cleanSchema(subtree.getRoot());
        } else
            if (root instanceof IfExpr) {
                XATTree thenTree = ((IfExpr) root).getThenXAT();
                cleanSchema(thenTree.getRoot());
                XATTree elseTree = ((IfExpr) root).getElseXAT();
                cleanSchema(elseTree.getRoot());
            }
}
/**
 * Delete redundent Rename operator
 * This method compares the old column name and new column, if they are equal
 * or both of them are null, then return true, else return false.
 * @return boolean true: means can be deleted, because the new = old, otherwise, cannot be deleted.
 * @param ncOp edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn
 * @since 1.0   
 * @see   edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.NameColumn
 */

protected boolean deleteRedundantRename(NameColumn ncOp) {
    if (ncOp.getNewColumnName() == null && ncOp.getOriginalColumnName() == null)
        return true;
    return ncOp.getNewColumnName().equals(ncOp.getOriginalColumnName());
}
/**
 * Find the nodes matching the give NavigationSteps.
 * This method traverse a DOMPattern starting from the root and returns all 
 * nodes that match the NavigationSteps. The Vector toFill contains all of 
 * the matches found. This is used to aid in Navigate / Tagger cancelout by 
 * finding all DOMPatternNodes that match the Navigate node.
 *
 * <b> Logic for determining a match </b>
 * <li>A ColumnNameNode will be a match if the tagValue if the last Step is toMatch
 * and the ColumnNameNode's parent's tagName corresponds to 2nd to last toMatch step
 * toMatch array</li>
 * <li> An AttributeNode will be a match if the last toMatch step begins with "@"
 * and the text after that corresponds to the tagName of the node </li>
 * <li> A TagNode will be a match if the tagName corresponds to any value in the
 * toMatch array </li>
 * <li>A TextNode could not be canceled out </li>
 * </pre>
 * @return void 
 * @param root The root of the DOMPatternNode.
 * @param toMatch The NavigationStep used as match pattern.
 * @param toFill All the matches found.
 * @since 1.0   
 * @see DOMPattern
 * @see DOMPatternNode
 * @see ColumnNameNodeImp
 * @see AttributeNodeImp
 * @see TagNodeImp
 * @see TextNodeImp
 */
protected void findMatchingPatternNode(
    DOMPatternNode root,
    NavigationStep[] toMatch,
    Vector toFill) {

    DOMPatternNode[] childrenRoot = root.getChildren();
    DOMPatternNode currentNode;

    //iterate over all children of the root
    for (int i = 0; i < childrenRoot.length; i++) {
        currentNode = childrenRoot[i];
        /**
        *  Look for a matching node, the strategy depends
        *  on the type of DOMPatternNode that we are looking at
        */

        /**
        * A TagNode will be a match if the tagName corresponds to any value in the
        * toMatch array
        */
        if (currentNode instanceof TagNode) {

            for (int x = 0; x < toMatch.length; x++) {
                if (toMatch[x].getStepValue().equals(currentNode.getTagName())) {
                    toFill.addElement(currentNode);
                }
                break;

            }

        }

        /**
         * ColumnNameNodes wont match b/c the user query doesnt know what the name of
         * the columns are, but only the names of the tags.  
         */
        if (currentNode instanceof ColumnNameNode) {
        }

        /**
        * An AttributeNode will be a match if the last toMatch step begins with "@"
        * and the text after that corresponds to the tagName of the node
        * 
        */
        if (currentNode instanceof AttributeNode) {
            int length = toMatch.length;
            //check for "@"
            String lastStep = toMatch[length - 1].getStepValue();
            if (lastStep.startsWith("@")) {
                //drop the "@"
                lastStep = lastStep.substring(1);
                if (lastStep.equals(currentNode.getTagName())) {
                    toFill.addElement(currentNode);
                }

            }

        }

        /**************
        ***** A TextNode will be a match if......
        *****   not really sure
        */
        if (currentNode instanceof TextNode) {

        }

        //loop recursively until there are no children of a node
        if (childrenRoot.length == 0) {
            //do nothing		
        } else {
            findMatchingPatternNode(childrenRoot[i], toMatch, toFill);
        }

    }

}
/**
 * Change all attached sub tree output column names.
 * This method change all the oldColumn name as newColumn name. 
 * Call recursively if the child node is of type GroupBy.
 * @return void
 * @param The XATTree attached to a GroupBy node (GroupBy, Function, Aggregate).
 * @param oldColumn The name of old column.
 * @param newColumn The name of new column.
 * @since 1.0   
 * @see GroupBy
 * @see XATOperator
 * @see XATTree.
 */
protected void nameColumnSubTree(
    XATOperator child,
    String oldColumn,
    String newColumn) {
    if (child.getOutputColumnName() != null) {
        if (child.getOutputColumnName().equals(oldColumn)) {
            child.setOutputColumnName(newColumn);
        }

        if (child instanceof GroupBy) {
            GroupBy groupChild = (GroupBy) child;
            nameColumnSubTree(
                groupChild.getXATTree().getRoot().getOperator(),
                oldColumn,
                newColumn);
        }

    }
    return;
}
/**
 * The Method recursively looks at the left and right sides of an Expression.
 * Traverses through the Expression tree until it finds all Terminal Expressions.
 * It then checks for if the TE's column equals the column to rename.
 * If so, replace all the old column name as new column name.
 * @return void
 * @param exp The expression going to be searched.
 * @param toMatch The name of the column being searched.
 * @param newName The new column name.
 * @since 1.0   
 * @see BinExpression
 * @see TerminalExpression
 * @see ColumnNameNodeImp
 */
protected void renameExpression(
    BinExpression exp,
    String newName,
    String toMatch) {

    //check for a terminal expression
    if (exp.getLeft() instanceof TerminalExpression) {
        TerminalExpression te1 = (TerminalExpression) exp.getLeft();
        //check for a column name terminal expression
        if (te1.getType() == te1.NAME) {
            String columnValue = ((ColumnName) te1.getValue()).getColumnName();
            //if there is a match
            if (columnValue.equals(toMatch)) {
                ColumnName cn1 = new ColumnNameImp(newName);
                te1.setValue(cn1);
                //updateVariableTable(newName, columnValue);
            }
        }
        te1 = (TerminalExpression) exp.getRight();
        //check for a column name terminal expression
        if (te1.getType() == te1.NAME) {
            String value = ((ColumnName) te1.getValue()).getColumnName();
            //if there is a match
            if (value.equals(toMatch)) {
                ColumnName cn1 = new ColumnNameImp(newName);
                te1.setValue(cn1);
                getVariableTable().updateVariableTable(newName, value);
            }
        }

    } else {
        try {
            renameExpression((BinExpression) exp.getLeft(), newName, toMatch);
            renameExpression((BinExpression) exp.getRight(), newName, toMatch);
        } catch (ClassCastException e) {
            System.out.println("Left Class name: " + exp.getLeft().getClass().getName());
            System.out.println("Right Class name: " + exp.getRight().getClass().getName());
        }

    }

}
/**
 * This method recursively look at the DOM Tree, traverses through the DOMPattern tree, 
 * which has a DOMPatternNoe as root, until it finds all TagNode and ColumnNameNode.
 * It then checks for if the node's column equals the column to rename. 
 * If so, for tag node, replace the old tag name as new tag name, 
 * for ColumnNameNode, then replace the ole TagValue as newName.
 * @return void
 * @param root The root node of DOM pattern tree.
 * @param newName The new column name.
 * @param toMatch The name of the column being searhed.
 * @since 1.0   
 * @see DOMPattern
 * @see DOMPatternNode
 * @see ColumnNameNodeImp
 * @see TagNodeImp
 */
protected void renameTagger(
    DOMPatternNode root,
    String toMatch,
    String newName) {

    DOMPatternNode[] childrenRoot = root.getChildren();
    DOMPatternNode currentNode;

    //iterate over all children of the root
    for (int i = 0; i < childrenRoot.length; i++) {
        currentNode = childrenRoot[i];
        if (currentNode instanceof TagNode) {

            if (toMatch.equals(currentNode.getTagName())) {
                currentNode.setTagName(newName);
            }
        }

        if (currentNode instanceof ColumnNameNode) {
            if (toMatch.equals(currentNode.getTagValue())) {
                //create a NavigationStep array corresponding to the newName
                VarStep v1 = new VarStepImp(newName);
                NavigationStep[] n1 = new NavigationStep[1];
                n1[0] = v1;
                currentNode.setTagValue(n1);
            }
        }

        //loop recursively until there are no children of a node
        if (childrenRoot.length == 0) {
            //do nothing		
        } else {
            renameTagger(childrenRoot[i], toMatch, newName);
        }
    }
}
/**
 * Rules for pushing a NameColumn operator through a Function operator.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Function).
 * @since 1.0   
 * @see NameColumn
 * @see Function
 * @see ReWriteResult
 */
public Object visit(NameColumn parent, Function child) {
    if (isDebug()) {
        System.out.println("***********In NameColumn, Function*********");
    }
    ReWriteResult p1 = new ReWriteResult();

    String oldColumn = parent.getOriginalColumnName();
    String newColumn = parent.getNewColumnName();

    if (child.getOutputColumnName().equals(oldColumn)) {
        //if the table contains the old value, update it
        if (this.getVariableTable().contains(oldColumn))
            getVariableTable().updateVariableTable(newColumn, child.getOutputColumnName());
        child.setOutputColumnName(newColumn);
    }

    Object[] params = child.getParams();
    boolean update = false;
    for (int i = 0; i < params.length; i++) {
        if (params[i] instanceof String) {
            String p = (String) params[i];
            if (p.equals(oldColumn)) {
                params[i] = newColumn;
                update = true;
            }
        }
    }
    //if any item in the param list was changed, update the param list
    if (update)
        child.setParams(params);

    //Swap and Go
    p1.setSwapped(true);
    p1.setFinalPosition(false);
    swap(parent, child);

    return p1;
}
/**
 * Rules for pushing a NameColumn node through a Source Node. 
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Source).
 * @since 1.0   
 * @see NameColumn
 * @see Source
 * @see ReWriteResult
 */
public Object visit(NameColumn parent, Source child) {

    ReWriteResult p1 = new ReWriteResult();
    if (isDebug()) {
        System.out.println("***********In NameColumn, Source*********");
    }
    String oldColumnName = parent.getOriginalColumnName();
    String newColumnName = parent.getNewColumnName();
    String sourceOutput = child.getOutputColumnName();
    //if the old column name equals the output of the source, then rename it
    if (oldColumnName.equals(sourceOutput)) {
        //if the table contains the old value, update it
        if (this.getVariableTable().contains(oldColumnName))
            getVariableTable().updateVariableTable(
                newColumnName,
                child.getOutputColumnName());
        child.setOutputColumnName(newColumnName);
    }

    //System.out.println(parent.toString() + " is now pushedDown");
    p1.setFinalPosition(true);
    p1.setSwapped(false);
    //and we will delete the NameColumn
    p1.setDeleteParentNode(true);

    return p1;
}
/**
 * Rules for pushing a NameColumn node through a generic XATOperator 
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (CartesianProduct).
 * @since 1.0   
 * @see NameColumn
 * @see CartesianProduct
 * @see ReWriteResult
 * @see renameExpression(BinExpression, String, String) renameExpression
 * @see deleteRedundantRename(NameColumn) deleteRedundantRename
 */
public Object visit(NameColumn parent, CartesianProduct child) {
    if (isDebug()) {
        System.out.println(
            "***********In NameColumn, CartesianProduct (Join)*********");
    }

    ReWriteResult p1 = new ReWriteResult();

    if (this.deleteRedundantRename(parent)) {
        p1.setDeleteParentNode(true);
        p1.setFinalPosition(true);
        if (isDebug()) {
            System.out.println("Redundant " + parent + " to be deleted!");
        }
        return p1;
    }

    try {
        String oldColumn = parent.getOriginalColumnName();
        String newColumn = parent.getNewColumnName();

        //now check the outputColumn
        if (child.getOutputColumnName() != null) {
            if (child.getOutputColumnName().equals(oldColumn)) {
                // update binding table to replace oldColumn into newColumn
                if (this.getVariableTable().contains(oldColumn))
                    this.getVariableTable().updateVariableTable(newColumn, oldColumn);
                child.setOutputColumnName(newColumn);

            }
        }
        //} else { // if the output name doesn't match, let's just swap it.
        // check if the operator consumed the original name used in the name column operator.
        ConsumedVisitor Vconsumed = new ConsumedVisitor();

        HashSet consumed = (HashSet) Vconsumed.visit(child);

        // if no columns are consumed, as in LOJ with no condition, or various MultiSource
        //  operators, skip this region of code.
        if (consumed != null) {
            if (consumed.contains(parent.getOriginalColumnName())) {

                // update the column name in the consumed place.
                if (child instanceof Join) {
                    renameExpression(
                        (BinExpression) ((Join) child).getExpression(),
                        oldColumn,
                        newColumn);
                } else {
                    if (isDebug()) {
                        System.err.println("WARNING!!! Unknown Operator to swap!" + child);
                    }
                }

            }
        }

        //push to correct branch of multiple source operator.
        // iterate through the subtrees and figure out the schema of the root
        // to check who nees this name operator.
        XATNode[] children = child.getNode().getChildren();
        boolean moved = false;
        for (int i = 0; i < children.length; i++) {
            // After using the Schema visitor, we are going to have the schema of that sub tree identified by the root node.
            CompleteSchemaVisitor schemaV = new CompleteSchemaVisitor();
            schemaV.visit(children[i]);
            // Now, let's get the schema of the children[i].
            Set schema = children[i].getSchema();
            // clean the schema
            cleanSchema(children[i]);

            // If the hashset has that original name, we push the name operator down.
            // we might push the name operator into multiple subtrees if it appeared
            // in those subtrees, because decorrelation might clone the subtrees.
            if (schema.contains(oldColumn)) {
                moved = true;
                // let's copy the rename node under
                XATNode cloneRename = (XATNode) parent.getNode().clone();
                // Inserted this cloned rename into the subtree.
                // basically, insert the clone between the child and children[i].
                // child is the "multi-source-operator", and children[i] is subtree of the child.
                children[i].setParent(cloneRename);
                cloneRename.setParent(child.getNode());
                child.getNode().setChild(i, cloneRename);
                if (cloneRename.getChildrenSize() != 0)
                    cloneRename.setChild(0, children[i]);
                else
                    cloneRename.addChild(children[i]);
                if (isDebug()) {
                    System.out.println(
                        "Copied rename operator to the branch " + children[i] + " of " + child);
                }
            }
        }

        // delete the rename operator in all the cases.
        // if the rename operator is not used anywhere below, then it's useless.
        p1.setDeleteParentNode(true);
        p1.setFinalPosition(true);
        if (isDebug()) {
            System.out.println(parent + " to be deleted!");
        }
        return p1;

    } catch (Exception e) {
        e.printStackTrace(System.err);
    }

    return null;

}
/**
 * Rules for pushing a NameColumn node through a GroupBy 
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (XATOperator).
 * @since 1.0   
 * @see NameColumn
 * @see Function
 * @see ReWriteResult
 * @see nameColumnSubTree(XATOperator, String, String) nameColumnSubTree
 */
public Object visit(NameColumn parent, GroupBy child) {
    if (isDebug()) {
        System.out.println("***********In NameColumn, GroupBy*********");
    }
    ReWriteResult p1 = new ReWriteResult();

    String oldColumn = parent.getOriginalColumnName();
    String newColumn = parent.getNewColumnName();

    //now check the outputColumn
    if (child.getOutputColumnName() != null) {
        if (child.getOutputColumnName().equals(oldColumn)) {
            //if the table contains the old value, update it
            if (this.getVariableTable().contains(oldColumn))
                getVariableTable().updateVariableTable(newColumn, child.getOutputColumnName());
            child.setOutputColumnName(newColumn);
        }
    }

    ((ColumnNameList) child).updateColumnName(oldColumn, newColumn);

    nameColumnSubTree(
        child.getXATTree().getRoot().getOperator(),
        oldColumn,
        newColumn);
    p1.setFinalPosition(false);
    p1.setSwapped(true);
    swap(parent, child);
    if (isDebug()) {
        System.out.println("Found GroupBy, fixed SubTree, swapped");
    }

    return p1;

}
/**
 * Rules for pushing a NameColumn operator through a Select operator.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Select).
 * @since 1.0   
 * @see NameColumn
 * @see Select
 * @see ReWriteResult
 * @see renameExpression(BinExpression, String, String) renameExpression
 */
public Object visit(NameColumn parent, Select child) {
    if (isDebug()) {
        System.out.println("***********In NameColumn, Select*********");
    }
    ReWriteResult p1 = new ReWriteResult();

    String oldColumn = parent.getOriginalColumnName();
    String newColumn = parent.getNewColumnName();

    if (child.getOutputColumnName().equals(oldColumn)) {
        //if the table contains the old value, update it
        if (this.getVariableTable().contains(oldColumn))
            getVariableTable().updateVariableTable(newColumn, child.getOutputColumnName());
        child.setOutputColumnName(newColumn);
    }

    ConsumedVisitor Vconsumed = new ConsumedVisitor();

    HashSet consumed = (HashSet) Vconsumed.visit(child);

    // if no columns are consumed, as in LOJ with no condition, or various MultiSource
    //  operators, skip this region of code.
    if (consumed != null) {
        if (consumed.contains(parent.getOriginalColumnName())) {
            renameExpression(
                (BinExpression) ((Select) child).getExpression(),
                oldColumn,
                newColumn);

        }
    }
    p1.setSwapped(true);
    p1.setFinalPosition(false);
    swap(parent, child);

    return p1;
}
/**
 * Rules for pushing a NameColumn node through a generic XATOperator.
 * MultipleSource are operators exclude CartesianProduct family operators.
 * For example: Merge, Set operator, IF_THEN_ELSE operator, "Or" operator.
 * In this case, for the semantic will be changed, no push down will be applied. 
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (XATMultiSourceOperator).
 * @since 1.0   
 * @see NameColumn
 * @see XATMultiSourceOperator
 * @see ReWriteResult
 * @see deleteRedundantRename(NameColumn) deleteRedundantRename 
 */
public Object visit(NameColumn parent, XATMultiSourceOperator child) {
    if (child instanceof CartesianProduct)
        return visit(parent, (CartesianProduct) child);

    if (isDebug()) {
        System.out.println("***********In NameColumn, XATMultiSourceOperator*********");
    }

    ReWriteResult p1 = new ReWriteResult();

    if (this.deleteRedundantRename(parent)) {
        p1.setDeleteParentNode(true);
        p1.setFinalPosition(true);
        if (isDebug()) {
            System.out.println("Redundant " + parent + " to be deleted!");
        }
        return p1;
    }

    // don't do anything else.
    p1.setFinalPosition(true);
    p1.setSwapped(false);
    return p1;
}
/**
 * Rules for pushing a NameColumn operator through an Aggregate operator.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Aggregate).
 * @since 1.0   
 * @see NameColumn
 * @see Aggregate
 * @see ReWriteResult
 */
public Object visit(NameColumn parent, Aggregate child) {
    if (isDebug()) {
        System.out.println("***********In NameColumn, Select*********");
    }
    ReWriteResult p1 = new ReWriteResult();

    String oldColumn = parent.getOriginalColumnName();
    String newColumn = parent.getNewColumnName();

    if (child.getOutputColumnName().equals(oldColumn)) {
        //if the table contains the old value, update it
        if (this.getVariableTable().contains(oldColumn))
            getVariableTable().updateVariableTable(newColumn, child.getOutputColumnName());
        child.setOutputColumnName(newColumn);
    }

    p1.setSwapped(true);
    p1.setFinalPosition(false);
    swap(parent, child);

    return p1;
}
/**
 * Rules for pushing a NameColumn node through a Navigate Node.
 * If the Navigate's binding, entryPoint or destination matches
 * the oldColumnName from the NameColumn, then change it to the newColumnName.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Navigate).
 * @since 1.0   
 * @see NameColumn
 * @see Function
 * @see ReWriteResult
 * @see deleteRedundantRename(NameColumn) deleteRedundantRename 
 */
public Object visit(NameColumn parent, Navigate child) {

    if (isDebug()) {
        System.out.println("***********In NameColumn, Navigate*********");
    }

    ReWriteResult p1 = new ReWriteResult();

    if (this.deleteRedundantRename(parent)) {
        p1.setDeleteParentNode(true);
        p1.setFinalPosition(true);
        if (isDebug()) {
            System.out.println("NameColumn" + parent + " to be deleted!");
        }
        return p1;
    }

    String oldName = parent.getOriginalColumnName();
    String newName = parent.getNewColumnName();
    //check the binding of the Navigate
    //if the binding equals the oldName, then update and delete the
    //	rename column, as no other node below it will use the oldName
    if (child.getBinding().equals(oldName)) {
        //if the table contains the old value, update it
        if (!this.isUsedAbove(parent.getNode(), oldName)) {

            if (this.getVariableTable().contains(oldName))
                getVariableTable().updateVariableTable(newName, child.getBinding());
            child.setBinding(newName);
            child.setOutputColumnName(newName);

            //Changed here XXX to delete NameColumn as the OldColumn
            //	won't be used by any other operators under it for it was
            //	created by this Navigate operator
            p1.setFinalPosition(true);
            p1.setSwapped(false);
            p1.setDeleteParentNode(true);
            return p1;
        }
    }
    //check the entryPoint
    if (!this.isUsedAbove(parent.getNode(), oldName)) {
        //try {
        NavigationStep[] entryPoint = child.getEntryPoint();
        for (int i = 0; i < entryPoint.length; i++) {
            if (entryPoint[i].getStepValue().equals(oldName)) {
                //System.out.println("Old value: " + entryPoint[i].getStepValue() + " new value: " + newName);
                entryPoint[i].setStepValue(newName);
                NavigationStep[] childDestination = child.getDestination();
                NavigationStep[] combinedChild = combineSteps(entryPoint, childDestination);
                getVariableTable().updateVariableTable(child.getBinding(), entryPoint);
            }
        }
        //} catch (NullPointerException e1) {
        //do nothing
        //}
        p1.setFinalPosition(false);
        p1.setSwapped(true);
        swap(parent, child);
        return p1;
    }
    p1.setFinalPosition(true);
    p1.setSwapped(false);
    return p1;
}
/**
 * Rules for pushing a NameColumn operator through a Tagger operator.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (Tagger).
 * @since 1.0   
 * @see NameColumn
 * @see Tagger
 * @see ReWriteResult
 */
public Object visit(NameColumn parent, Tagger child) {
    if (isDebug()) {
        System.out.println("***********In NameColumn, Tagger*********");
    }
    ReWriteResult p1 = new ReWriteResult();

    String oldColumn = parent.getOriginalColumnName();
    String newColumn = parent.getNewColumnName();

    renameTagger(((DOMPattern) child.getPattern()).getRoot(), oldColumn, newColumn);
    //now check the outputColumn
    if (child.getOutputColumnName().equals(oldColumn)) {
        //if the table contains the old value, update it
        if (this.getVariableTable().contains(oldColumn))
            getVariableTable().updateVariableTable(newColumn, child.getOutputColumnName());
        child.setOutputColumnName(newColumn);
    }
    p1.setSwapped(true);
    p1.setFinalPosition(false);
    swap(parent, child);

    return p1;
}
/**
 * <pre>
 * Select MultipleSource Operator Product Rule.
 * we want to determine what to do with the Select,  there are 4 possiblities.
 * 1) The Select should move to the left branch
 * 2) The Select should move to the right branch
 * 3) The Select should be merged with the Cartesian to form an OuterJoin
 * 4) The Select should stay on top of the Cartesian
 * to determine which case we have, we will try to findMatchingNavigateNode on both the
 * left and right branches of the Cartesian with both the left and right column names of the select
 * to accomplish this, we will use a state machine
 * two booleans, left and right are set to true.  if any column name does not go to that side,
 * then the boolean is set to false
 * after all column names have been checked against all branches, we can determine which of the 4
 * actions to take
 * 1) true / false
 * 2) false / true
 * 3) true / true
 * 4) false / false
 * </pre>
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Select).
 * @param child child XAT Operator (CartesianProduct).
 * @since 1.0   
 * @see Select
 * @see CartesianProduct
 * @see ReWriteResult
 * @see findMatchingNavigationNode(String, XATNode) findMatchingNavigationNode
 * @see moveNode(XATNode, XATNode) moveNode
 */
public Object visit(Select parent, CartesianProduct child) {
    if (isDebug()) {
        System.out.println("In Select / CartesianProduct (Join)");
    }
    ReWriteResult r1 = new ReWriteResult();
    //the bindings = null when they are not column names
    /*
    String leftBinding = getLeftColumn((BinExpression) parent.getExpression());
    String rightBinding = getRightColumn((BinExpression) parent.getExpression());
    */
    Vector columnNames = new Vector();
    getColumnNames((BinExpression) parent.getExpression(), columnNames);
    Vector joinBranches = new Vector();
    boolean left = false;
    boolean right = false;
    //get all of the children of Cartesian
    getChildren(joinBranches, (XATNode) child.getNode());
    for (int i = 0; i < columnNames.size(); i++) {
        /*we check only if the left still is true */
        if (!left) {
            /*if we cant find a matching node on the left, set it to false so we dont look again */
            if (findMatchingNavigationNode((String) columnNames.elementAt(i),
                (XATNode) joinBranches.elementAt(0))) {
                left = true;
            }
        }
        if (!right) {
            if (findMatchingNavigationNode((String) columnNames.elementAt(i),
                (XATNode) joinBranches.elementAt(1))) {
                right = true;
            }
        }

    }

    //now determine what to do
    //both are false so keep Select where it is
    if (!left && !right) {
        r1.setFinalPosition(true);
        r1.setSwapped(false);

    }
    //left = true, right = false so move to the left side
    if (left && !right) {
        moveNode((XATNode) parent.getNode(), (XATNode) joinBranches.elementAt(0));
        if (isDebug()) {
            System.out.println("^^^^^^^^^^^^^ Moved Select to the left side of the MSO");
        }
        r1.setSwapped(false);
        r1.setMoved(true);
        r1.setFinalPosition(false);

    }
    //both = true so create a join
    if (left && right) {
        if (child instanceof Join) {
            // if it is already a join
            // do nothing.
            r1.setSwapped(false);
            r1.setFinalPosition(true);
        } else
            if (child instanceof CartesianProduct) {
                XATNode x1 = child.getNode();
                if (isDebug()) {
                    System.out.println("^^^^^^^^^^^^^ Created a Join");
                }
                Join j1 = new ThetaJoinImp(x1, (BinExpression) parent.getExpression());
                x1.setOperator(j1);
                //delete the parent
                r1.setDeleteParentNode(true);
                //the select is in its final position
                r1.setFinalPosition(true);
            }

    }
    /* 
    * the only case left is left = false, right = true
    */
    if (!left && right) {
        moveNode((XATNode) parent.getNode(), (XATNode) joinBranches.elementAt(1));
        if (isDebug()) {
            System.out.println(
                "^^^^^^^^^^^^^ Moved Select to the right side of the XATMultiSourceOperator");
        }
        r1.setSwapped(false);
        r1.setMoved(true);
        r1.setFinalPosition(false);

    }
    return r1;
}
/**
 * Select GroupBy Rule.  Select cannot be pushed past the Groupby
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Select).
 * @param child child XAT Operator (GroupBy).
 * @since 1.0   
 * @see Select
 * @see GroupBy
 * @see ReWriteResult
 */
public Object visit(Select parent, GroupBy child) {
    if (isDebug()) {
        System.out.println("************In Select / GroupBy *********");
    }
    ReWriteResult r1 = new ReWriteResult();
    r1.setFinalPosition(true);
    r1.setSwapped(false);
    return r1;
}
/**
 * <pre>
 * Select MultipleSource Operator Product Rule.
 * MultipleSource are operators exclude CartesianProduct family operators.
 * For example: Merge, Set operator, IF_THEN_ELSE operator, "Or" operator.
 *
 * In this case, no push down will be applied. For the semantic will be changed.
 * </pre>
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Select).
 * @param child child XAT Operator (XATMultiSourceOperator).
 * @since 1.0   
 * @see Select
 * @see XATMultiSourceOperator
 * @see ReWriteResult
 */
public Object visit(Select parent, XATMultiSourceOperator child) {
    if (child instanceof CartesianProduct)
        return visit(parent, (CartesianProduct) child);

    if (isDebug()) {
        System.out.println("In Select / XATMultiSourceOperator");
    }

    // stop the pushdown of the select.

    ReWriteResult r1 = new ReWriteResult();
    r1.setFinalPosition(true);
    r1.setSwapped(false);

    return r1;
}
/**
 * Select XATOperator Rule.
 * Select can always be swapped with any XATOperator that doesnt have its
 * own rule.
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Select).
 * @param child child XAT Operator (XATOperator).
 * @since 1.0   
 * @see Select
 * @see XATOperator
 * @see ReWriteResult
 */
public Object visit(Select parent, XATOperator child) {
    if (isDebug()) {
        System.out.println("In Select / XATOperator");
    }
    ReWriteResult r1 = new ReWriteResult();
    // if the select depends on the child operator.

    if (child.getOutputColumnName() != null) {
        if (parent.getExpression().toString().indexOf(child.getOutputColumnName())
            != -1) {
            // stop here.
            r1.setFinalPosition(true);
            r1.setSwapped(false);
        } else {
            // keep going.
            swap(parent, child);
            r1.setFinalPosition(false);
            r1.setSwapped(true);
        }
    } else {
        //the child operator does not generate anything, parent could not dependent on it
        swap(parent, child);
        r1.setFinalPosition(false);
        r1.setSwapped(true);
    }
    return r1;
}
/**
 * Select Navigate Rule.
 * Select can swap with a Navigate if the bindings dont match
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Select).
 * @param child child XAT Operator (Navigate).
 * @since 1.0   
 * @see Select
 * @see Navigate
 * @see ReWriteResult
 */
public Object visit(Select parent, Navigate child) {
    if (isDebug()) {
        System.out.println("************In Select / Navigate *********");
    }
    ReWriteResult r1 = new ReWriteResult();

    boolean match = false;
    String navigateBinding = child.getBinding();
    //String selectBinding = getColumnName((BinExpression)parent.getExpression());
    ColumnName[] selectBindings =
        ((BinExpression) parent.getExpression()).getRelatedColumnNames();
    for (int i = 0; i < selectBindings.length; i++)
        if (selectBindings[i].getColumnName().equals(navigateBinding)) {
            match = true;
            break;
        }

    if (match) {
        r1.setFinalPosition(true);
        r1.setSwapped(false);
    } else {
        swap(parent, child);
        r1.setSwapped(true);
        r1.setFinalPosition(false);
    }

    return r1;
}
/**
 * Rules for removing a useless Aggregation operator.
 * @return ReWriteResult. The swap flag in ReWriteResult true if the nodes 
 * were swapped, false otherwise.
 * @param parent parent XAT operator (Aggregate).
 * @param child child XAT Operator.
 * @since 1.0   
 * @see XATOperator
 * @see ReWriteResult
 */
public Object visit(Aggregate parent, XATOperator child) {
    if (isDebug()) {
        System.out.println("In Aggregate / XATOperator: " + child.toString());
    }
    ReWriteResult p1 = new ReWriteResult();
    // get output column name of this operator.
    String outCol = parent.getOutputColumnName();
    // if not used above and it is not a root, cut it
    if (parent.getNode().getParent() != null) {
        if (!isUsedAbove(parent.getNode(), outCol)) {
            if (isDebug()) {
                System.out.println(parent + " to be deleted!");
            }
            p1.setDeleteParentNode(true);
        }
    }
    p1.setFinalPosition(true);
    return p1;

}
/**
 * Rules for pushing a Navigate operator through a NameColumn operator
 * A Navigate node can always be pushed through a NameColumn operartor
 * This rule is the same as NameColumn / Navigate
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (NameColumn).
 * @since 1.0   
 * @see Navigate
 * @see GroupBy
 * @see ReWriteResult
 * @see deleteRedundantRename(NameColumn) deleteRedundantRename	 
 */
public Object visit(Navigate parent, NameColumn child) {
    if (isDebug()) {
        System.out.println("************In Navigate / NameColumn *********");
    }
    ReWriteResult p1 = new ReWriteResult();
    if (this.deleteRedundantRename(child)) {
        p1.setDeleteChildNode(true);
        p1.setFinalPosition(false);
        if (isDebug()) {
            System.out.println("NameColumn" + child + " to be deleted!");
        }
        return p1;
    }

    String oldName = child.getOriginalColumnName();
    String newName = child.getNewColumnName();
    //check the binding of the Navigate
    if (parent.getBinding().equals(oldName)) {
        //if the old name exists in the table, update it.
        if (this.getVariableTable().contains(oldName))
            getVariableTable().updateVariableTable(newName, parent.getBinding());
        parent.setBinding(newName);
        parent.setOutputColumnName(newName);

        //Changed here XXX to delete NameColumn as the OldColumn
        //	won't be used by any other operators under it for it was
        //	created by this Navigate operator
        p1.setFinalPosition(true);
        p1.setSwapped(false);
        p1.setDeleteChildNode(true);
        return p1;
    }
    //check the entryPoint
    NavigationStep[] entryPoint = parent.getEntryPoint();
    for (int i = 0; i < entryPoint.length; i++) {
        if (entryPoint[i].getStepValue().equals(oldName)) {
            if (isDebug()) {
                System.out.println(
                    "Old value: " + entryPoint[i].getStepValue() + " new value: " + newName);
            }
            entryPoint[i].setStepValue(newName);
            NavigationStep[] parentDestination = parent.getDestination();
            NavigationStep[] combinedParent = combineSteps(entryPoint, parentDestination);
            getVariableTable().updateVariableTable(parent.getBinding(), entryPoint);
        }
    }

    if (!child
        .getOutputColumnName()
        .equals(parent.getEntryPoint()[0].getStepValue())) {
        p1.setFinalPosition(false);
        p1.setSwapped(true);
        swap(parent, child);
    } else {
        p1.setFinalPosition(true);
    }
    return p1;

}
/**
 * Rules for pushing a Navigate operator through a Source operator
 * A Navigate can never go through a Source operator.  The Source operators
 * are on the very bottom of trees and swapping the 2 would cause errors
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (Source).
 * @since 1.0   
 * @see Navigate
 * @see Source
 * @see ReWriteResult
 */
public Object visit(Navigate parent, Source child) {

    ReWriteResult p1 = new ReWriteResult();
    if (isDebug()) {
        System.out.println("***********In Navigate, Source*********");
    }
    //if the source node is not default.xml, the swap may be performed.
    //otherwise, the navigate node is as far down in the tree as possible
    p1.setFinalPosition(true);
    p1.setSwapped(false);
    return p1;
}
/**
 * <pre>
 * Rules for pushing a Navigate node through a XATMultiSourceOperator .
 * This is also the same rule for Navigate
 * The XATMultiSourceOperator node will have n branches, the Navigate node must be moved to
 * the branch that contains a Navigate node whose destination equals parent.entryPoint
 *
 * findMatchingNavigationNode is used to determine which branch to move to
 *
 * This rules apply to cartesian product operators.
 * </pre> 
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (NameColumn).
 * @param child child XAT Operator (XATMultiSourceOperator).
 * @since 1.0   
 * @see Navigate
 * @see XATMultiSourceOperator
 * @see ReWriteResult
 */
public Object visit(Navigate parent, CartesianProduct child) {
    if (isDebug()) {
        System.out.println("In Navigate / CartesianProduct (Join)");
    }
    ReWriteResult p1 = new ReWriteResult();
    //in case nothing is found, just put the Navigate in its final position
    p1.setSwapped(false);
    p1.setFinalPosition(true);

    Vector joinBranches = new Vector();
    Vector childrenVector = new Vector();
    XATNode currentNode;

    //get all of the children of join
    getChildren(joinBranches, (XATNode) child.getNode());

    //iterate over each child
    for (int x = 0; x < joinBranches.size(); x++) {
        //look at this branch for the navigate node
        VarStep v1 = (VarStep) parent.getEntryPoint()[0];
        // compute the full schema of this branch.
        CompleteSchemaVisitor completeSchema = new CompleteSchemaVisitor();
        Set schema = (Set) completeSchema.visit((XATNode) joinBranches.elementAt(x));
        // clean the schema
        cleanSchema((XATNode) joinBranches.elementAt(x));

        if (schema.contains(v1.getStepValue())) {
            //if the node was found, copy the navigate under that branch
            boolean result =
                copyNode((XATNode) parent.getNode(), (XATNode) joinBranches.elementAt(x));
            if (isDebug()) {
                System.out.println(
                    "I decided to copy the Navigate:"
                        + parent
                        + " to branch "
                        + x
                        + " of MSO:"
                        + child);
            }
            // delete the parent if it is copied.
            p1.setSwapped(false);
            p1.setDeleteParentNode(true);
            p1.setFinalPosition(true);
        }

    }

    return p1;
}
/**
 * Rules for pushing a Navigate operator through a Groupby operator
 * A Navigate can, but for the sake of order sensitive rewrite, is never
 *  pushed through a groupby (due to the way the tree is generated, this
 *  would normally only allow a navunnest of a text() to be pushed anyway).
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (GroupBy).
 * @since 1.0   
 * @see Navigate
 * @see GroupBy
 * @see ReWriteResult
 */
public Object visit(Navigate parent, GroupBy child) {

    ReWriteResult p1 = new ReWriteResult();
    if (isDebug()) {
        System.out.println("*********Navigate / GroupBy*********");
    }
    p1.setFinalPosition(true);
    p1.setSwapped(false);
    return p1;

}
/**
 * Rules for pushing a Navigate operator through a Project operator
 * A Navigate can always move through a Project, but the destination of the
 * Navigate node must be added to columns that are Projected
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (Project).
 * @since 1.0   
 * @see Navigate
 * @see Project
 * @see ReWriteResult
 * @depracated This rule should not happen in our expectation.
 */
public Object visit(Navigate parent, Project child) {

    ReWriteResult p1 = new ReWriteResult();
    p1.setFinalPosition(false);

    //the Navigate's destination must be added to the Project pattern
    if (parent.getBinding() == null) {
        String columnName = new String("");
        NavigationStep[] n1 = parent.getDestination();
        for (int i = 0; i < n1.length; i++) {
            columnName = columnName + n1[i] + "/";

        }
        child.insertColumnName(columnName, 0);
    } else {
        //if there is a variable binding, then use that instead of the destination
        child.insertColumnName(parent.getBinding(), 0);
    }
    //a Navigate and a Project can always be swapped
    swap(parent, child);
    p1.setSwapped(true);
    return p1;

}
/**
 * Rules for pushing a Navigate operator through a Select operator
 * A Navigate can always move through a Select
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (Select).
 * @since 1.0   
 * @see Navigate
 * @see Select
 * @see ReWriteResult
 */
public Object visit(Navigate parent, Select child) {

    ReWriteResult p1 = new ReWriteResult();
    if (isDebug()) {
        System.out.println("*********Navigate / Select*********");
    }
    //a navigate and a select can always be swapped
    swap(parent, child);
    p1.setFinalPosition(false);
    p1.setSwapped(true);
    return p1;

}
/**
 * <pre>
 * Rules for pushing a Navigate node through a XATMultiSourceOperator .
 * MultipleSource are operators exclude CartesianProduct family operators.
 * For example: Merge, Set operator, IF_THEN_ELSE operator, "Or" operator.
 *
 * In this case, no push down will be applied. For the semantic will be changed.
 * </pre> 
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (XATMultiSourceOperator).
 * @since 1.0   
 * @see Navigate
 * @see XATMultiSourceOperator
 * @see ReWriteResult
 */
public Object visit(Navigate parent, XATMultiSourceOperator child) {
    if (child instanceof CartesianProduct)
        return visit(parent, (CartesianProduct) child);

    if (isDebug()) {
        System.out.println("In Navigate / XATMultiSourceOperator Product");
    }
    ReWriteResult p1 = new ReWriteResult();
    //in case nothing is found, just put the Navigate in its final position
    p1.setSwapped(false);
    p1.setFinalPosition(true);

    return p1;
}
/**
 * Rules for pushing a Navigate operator through a generic XATOperator operator
 * A Navigate can always be pushed through a Distinct
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (XATOperator).
 * @since 1.0   
 * @see Navigate
 * @see XATOperator
 * @see ReWriteResult
 */
public Object visit(Navigate parent, XATOperator child) {
    if (isDebug()) {
        System.out.println("In Navigate / XATOperator: " + child.toString());
    }
    ReWriteResult p1 = new ReWriteResult();
    if (!child
        .getOutputColumnName()
        .equals(parent.getEntryPoint()[0].getStepValue())) {
        swap(parent, child);
        p1.setFinalPosition(false);
        p1.setSwapped(true);
    } else {
        p1.setFinalPosition(true);
        p1.setSwapped(false);
    }
    return p1;

}
/**
 * Rules for pushing a Navigate operator through a Navigate operator.
 * If the entry point for the parent is the same as the complete path for the child,
 * the nodes cannot be swapped and the parent is pushedDown.  Otherwise, the swap is legal.
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Navigate).
 * @param child child XAT Operator (Navigate).
 * @since 1.0   
 * @see Navigate
 * @see Navigate
 * @see ReWriteResult
 */
public Object visit(Navigate parent, Navigate child) {
    if (isDebug()) {
        System.out.println("************In Navigate / Navigate *********");
    }
    ReWriteResult p1 = new ReWriteResult();
    NavigationStep[] parentEntryPoint = parent.getEntryPoint();
    String parentEPString = parentEntryPoint[0].getStepValue();
    String childBinding = child.getBinding();
    /*
    NavigationStep[] childDestination = child.getDestination();
    NavigationStep[] childEntryPoint = child.getEntryPoint();
    NavigationStep[] combinedChild =
    combineSteps(childEntryPoint, childDestination);
    */
    try {

        //obtain a string representation of the destination and entry point

        //String childString = createString(combinedChild);

        //if they are the same, then they cannot be swapped
        if (!childBinding.equals(parentEPString)) {
            p1.setFinalPosition(false);
            swap(parent, child);
            p1.setSwapped(true);
        } else {
            p1.setFinalPosition(true);
            p1.setSwapped(false);

        }

    } catch (Exception e1) {
        System.out.println("Error!!! " + e1.toString());
        System.out.println("Parent: " + parent.toString());
        System.out.println("Child: " + child.toString());
        swap(parent, child);
        //System.out.println("done swapping nodes");
        p1.setSwapped(true);
        p1.setFinalPosition(true);
    }

    return p1;

}
/**
 * Rules for removing a useless Tagger operator 
 * @return ReWriteResult. The swap flag in ReWriteResult false.
 * @param parent parent XAT operator (Tagger).
 * @param child child XAT Operator (XATOperator).
 * @since 1.0   
 * @see Tagger
 * @see XATOperator
 * @see ReWriteResult
 */
public Object visit(Tagger parent, XATOperator child) {
    if (isDebug()) {
        System.out.println("In Tagger / XATOperator: " + child.toString());
    }
    ReWriteResult p1 = new ReWriteResult();
    // get output column name of this operator.
    String outCol = parent.getOutputColumnName();
    // if not used above and it is not a root, cut it
    if (parent.getNode().getParent() != null) {
        if (!isUsedAbove(parent.getNode(), outCol)) {
            if (isDebug()) {
                System.out.println(parent + " to be deleted!");
            }
            p1.setDeleteParentNode(true);
        }
    }
    p1.setFinalPosition(true);
    return p1;

}
}
