package edu.wpi.cs.dsrg.xmldb.xat.component.xattreegeneration;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.utils.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory.*;
import xacute.quilt.*;
import java.lang.reflect.Array;
import java.util.Vector;

/**
 * This class visits the XAT Tree constructed in ParserTraverse 
 * and get rid of For operaters. Thus we can get rid of inner loops
 * in the XAT Tree and make a flat algebra tree.
 * <p>
 * To understand this code, one should know how the decorrelation
 * works and what our algorithm is.  It would also be helpful to 
 * know what a visitor pattern is because that is what we use to
 * traverse through the query.
 * <p>
 * Currently, we need to 2 round of decorrelation to get rid of 
 * all For loops.
 * <p>
 * The Main issues in decorrelation is:
 * 1. empty collection problem
 * 2. count bugs.
 *
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern
 */
public class QueryDecorrelation extends InheritableVisitor {

	/**
    * record the context FOR bindings for the groupby operator generation.
    * 
    * @since 1.0
    */
	protected String[] contextForBindings; 
	
	/**
    * record the decorrelated XAT Tree.
    * 
    * @since 1.0
    */
	protected XATTree decorrTree;

	/**
    * A global point to denote where is the current parent we are working with.
    * 
    * @since 1.0
    */
	protected XATNode currentParent;

	/**
    * used to create operators
    * 
    * @since 1.0
    */
    protected OperatorBuilder builder ;
/**
 * Class Default Constructor
 *
 * @since 1.0
 */
protected QueryDecorrelation() {
    super();
}
/**
 * Class Constructor with outer For binding context
 *
 * @since 1.0
 */
protected QueryDecorrelation(String forContext[], XATTree tree) {
    //remember the FOR context from the caller.
    setForContext(forContext);
    //init the decorrelation.
    init(tree);
}
/**
 * Class Constructor with XAT Tree to be decorrelated.
 *
 * @since 1.0
 */
public QueryDecorrelation(XATTree tree) {
    // decorrelate it first time.
    init(tree);
    // decorrelate it second time to handle the empty sequence bug.
    // clear the for context.
    this.setForContext(new String[] {
    });
    init(this.decorrTree);
}
/**
 * Adding items to the FOR context, useful for nested Aggregates
 *
 * @param input String of for binding variable names.
 * @since 1.0
 */
private void addForContext(String[] input) {
    int lenFor = contextForBindings.length;
    int lenIn = input.length;

    // first store the input into a vector.
    Vector inV = new Vector();
    for (int i = 0; i < lenIn; i++) {
        inV.add(input[i]);
    }

    // first check if the new input is in this original context.
    for (int i = 0; i < inV.size(); i++) {
        // check if that's already there.
        for (int j = 0; j < lenFor; j++) {
            if (contextForBindings[j].equals(inV.elementAt(i))) {
                // remove it.
                inV.remove(i);
                i--;
            }
        }
    }

    String[] newForContext = new String[lenFor + inV.size()];

    for (int i = 0; i < lenFor; i++) {
        newForContext[i] = contextForBindings[i];
    }

    for (int i = 0; i < inV.size(); i++) {
        newForContext[lenFor + i] = (String) inV.elementAt(i);
    }

    setForContext(newForContext);
}
/**
 * Do complex decorrelation with LOJ.
 * <p>
 * The complex decorrelation includes two parts:
 * . A subtree with extra sources. --> LOJ will be created.
 * . A subtree without extra sources. --> Linear tree.
 * . To solve empty collection bug, we decide to generate LOJ for all FOR.
 * . even only one source exist.
 * 
 * @param outerTreeDecor QueryDecorrelation of outer For binding
 * @param forOp the For operator to be deleted.
 * @param subTreeDecor QueryDecorrelateion of the inner tree belong to this For operator.
 * @since 1.0
 */
void complexDecorrelation(
    QueryDecorrelation outerTreeDecor,
    FOR forOp,
    QueryDecorrelation subTreeDecor) {

    //create an OuterJoin node
    XATNode outerJoin = new XATNodeImp();
    LeftOuterJoin loj = builder.newLeftOuterJoin(outerJoin, null);
    loj.setOutputColumnName(forOp.getOutputColumnName());
    outerJoin.setOperator(loj);

    // we will get around the FOR node in order to delete it from the decorrelation tree.
    if (this.currentParent != null) { // if there is a parent of the FOR node,
        //now, insert the outerjoin node to the tree
        decorrTree.addNode(outerJoin, this.currentParent);
    } else { // otherwise, FOR is the root node. Let's make the outerJoinNode the root node.
        decorrTree.setRoot(outerJoin);
    }

    if (isDebug())
        System.out.println("Outer join added to tree...");

    //Then link the outquery to the left.
    decorrTree.addNode(outerTreeDecor.getDecorTree().getRoot(), outerJoin);

    //Then link the innerquery to the right.
    decorrTree.addNode(subTreeDecor.getDecorTree().getRoot(), outerJoin);

    // adding outerTreeDecor to all the leaves.
    // cartisian product will be generated if the leaf is a SOURCE
    // we need to make a clone of the outerquery and put those there.
    XATTree decorredSubTree = subTreeDecor.getDecorTree();
    XATNode[] leaves = decorredSubTree.getLeaves();
    if (leaves != null) {
        for (int ileaf = 0; ileaf < leaves.length; ileaf++) {

            //this checks for multiple sources, and adds a cartesian
            //product node if there are multiple sources
            //the 2 children of the cartesian product are the second source
            //and all the nodes beneath the first for node
            if (leaves[ileaf].getOperator() instanceof Source) {
                //if source's parent is already cartesian product,
                //skip the following statements.
                //xin@6/26/03
                if (leaves[ileaf].getParent().getOperator() instanceof CartesianProduct) {
                    continue;
                }
                //create a cartesian product node
                XATNode cpNode = new XATNodeImp();
                CartesianProduct cartProd = builder.newCartesianProduct(cpNode);
                cpNode.setOperator(cartProd);
                // now, add the cartesian product to the tree
                decorredSubTree.addNode(cpNode, leaves[ileaf].getParent());
                leaves[ileaf].getParent().deleteChild(leaves[ileaf]);
                //set the source as a child of the cartesian product
                decorredSubTree.addNode(leaves[ileaf], cpNode);

                //now, create a distinct node
                //this will also be set as a child of the cartesian product
                XATNode distinctNode = new XATNodeImp();
                Distinct dNode = builder.newDistinct(distinctNode);
                distinctNode.setOperator(dNode);
                dNode.setInputColumnName(forOp.getColumnNames()[0]);

                //add it to the tree
                decorredSubTree.addNode(distinctNode, cpNode);

                /*
                * put the outerquery below the distinctNode.
                */

                // clone the outer query.
                XATTree outerTreeClone = (XATTree) outerTreeDecor.getDecorTree().clone();

                // link the outer query.
                decorredSubTree.addNode(outerTreeClone.getRoot(), distinctNode);
            } else {

                XATTree outerTreeClone = (XATTree) outerTreeDecor.getDecorTree().clone();
                this.decorrTree.addNode(outerTreeClone.getRoot(), leaves[ileaf]);
            }
        }
    }

    // delete the forOp node.
    if (currentParent != null)
        currentParent.deleteChild(forOp.getNode());

}
/**
 * Create GroupBy operator at the place of an XAT operator and put the
 * XAT operator as the root of the inner tree of the new GroupBy operator.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 * @param op edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
 * @since 1.0
 */
XATOperator createGBOp(XATOperator op) {

    GroupBy gbOp = builder.newGroupBy(new XATNodeImp());
    gbOp.setOutputColumnName(op.getOutputColumnName());

    // set the groupBy's input columns.
    gbOp.setColumnNames(getForContext());

    // create the subtree of the operator.
    XATTree subTree = new XATTreeImp();

    // disconnect the op's node from it's children.
    XATNode opNode = (XATNode) op.getNode().clone();
    opNode.setChildren(new XATNodeImp[0]);

    // add the op's node to the subtree.
    subTree.setRoot(opNode);

    gbOp.setXATTree(subTree);

    return gbOp;
}
/**
 * Create GroupBy operator at the place of an XAT operator and put the
 * XAT operator as the root of the inner tree of the new GroupBy operator.
 * The inner tree of the new GroupBy operator will stop at the XAT operator
 * refered by the second parameter.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 * @param op edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
 * @since 1.0
 */
XATOperator createGBOp(XATOperator op,XATOperator subtreeEndOp) {

    GroupBy gbOp = builder.newGroupBy(new XATNodeImp());
    gbOp.setOutputColumnName(op.getOutputColumnName());

    // set the groupBy's input columns.
	gbOp.setColumnNames(getForContext());

	// create the subtree of the operator.
    XATTree subTree = new XATTreeImp();

    // disconnect the op's node from it's children.
	XATNode opOrig = op.getNode();
    XATNode opNode = (XATNode)opOrig.clone();
    opNode.setChildren(opOrig.getChildren());
    for (int i=0;i<opNode.getChildren().length;i++)
    	opNode.getChild(i).setParent(opNode);
    	
    XATNode subtreeEndOpNode = subtreeEndOp.getNode();
    subtreeEndOpNode.getChild(0).setParent(opOrig);
    opOrig.setChildren(subtreeEndOpNode.getChildren());
    
    subtreeEndOpNode.setChildren(new XATNodeImp[0]);
	op.setNode(opNode);
	opNode.setOperator(op);
	
    // add the op's node to the subtree.
    subTree.setRoot(opNode);

    gbOp.setXATTree(subTree);

    return gbOp;
}
/**
 * Completes the decorrelation algorithm on the XAT
 *
 * @param node XATNode: the root of the xat
 * @param parent XATNode: parent of the first parameter in XAT Tree.
 * @since 1.0
 */
void decorrelate(XATNode node, XATNode parent) {
    XATNode[] children = node.getChildren();
    this.currentParent = parent;

    visit(node.getOperator());

    // if the operator is FOR, no further decorrelation is needed.
    // because the decorrelation has been recursively called in
    // the visit(FOR) operator.
    if (node.getOperator() instanceof FOR)
        return;

    /**
    * In the future if visit(X) has the recursive
    * call of the QueryDecorrelation, then the following loop doesn't need
    * to be called again. But in above code, you need to check if you have
    * visited X.
    */

    for (int i = 0; i < children.length; i++) {
        decorrelate(children[i], node);
        this.currentParent = node;
    }
}
public Object dispatch() {
    //System.out.println(indent() + "NULL");
    return null;
}
/**
 * Called when an object hasn't been implemented
 * by the visitor pattern
 * default handling of the unknow operators.
 */
public Object dispatch(Object obj) {

    XATOperator opr = (XATOperator) obj;

    // do nothing here. Because we modify on the cloned input tree directly.	

    if (isDebug())
        System.out.println(
            "Operator of type " + opr.getClass().getName() + " is not handled!");
    if (isDebug())
        System.out.println("This node is simply added to the Decorrelated Tree as is.");

    return null;
}
/**
 * Access method of field decorrelate tree.
 *
 * @return XATTree
 * @since 1.0
 */
public XATTree getDecorTree() {
    return this.decorrTree;
}
/**
 * Access method of field context of outter for bindings.
 *
 * @return array of String
 * @since 1.0
 */
private String[] getForContext() {
    return contextForBindings;
}
/**
 * Initialize binding table and do decorrelation.
 *
 * @param inputXAT edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 * @since 1.0
 */
void init(XATTree inputXAT) {
    if (inputXAT != null) {
	    try {
	    	/** create the factory */
        	OperatorFactory factory = OperatorFactory.newInstance();
        	/** instantiate the operator builder */
        	builder = factory.newOperatorBuilder();
    	} catch (FactoryException e) {
        	/** 
        	* there was a problem instantiating the factory
        	* so we have to stop generating the tree. 
        	*/
        	System.err.println(e);
    	}
        decorrTree = (XATTree) inputXAT.clone();
        //start decorrelating
        decorrelate(decorrTree.getRoot(), null);
        //recompute the binding table.
        decorrTree.setBindingTable(new BindingTableComputer().computeBindingTable(decorrTree));
        //after decorrelation, cleanup the XATTree.
		if(System.getProperty("CLEANXAT")!=null)
                RewriteUtilities.cleanXATAfterGeneration(decorrTree);
    } else {
        decorrTree = new XATTreeImp();
        decorrTree.setBindingTable(new BindingTableComputer().computeBindingTable(decorrTree));
    }
}
/**
 * Get the value of the propertity 'DEBUG_QueryDecorrelation'.
 * @return boolean True: With debug information. False: No debug information.
 *
 * @since 1.0
 */
public boolean isDebug() {
	return System.getProperty("DEBUG_QueryDecorrelation")!=null;
}
/**
 * Find and move LOJ below specific node.
 *
 * @param xOp edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
 * @since 1.0
 */
protected void moveLOJ(XATOperator xOp) {

    XATOperator loj;
    for (loj = xOp.getNode().getParent().getOperator();
        !(loj instanceof LeftOuterJoin);
        ) {
        if (loj.getNode().getParent() == null) { //can not find LOJ till root.
            loj = null;
            break;
        }
        if (loj instanceof GroupBy) {
            //there is GB in the middle. cancel this kind of moving.
            return;
        }
        if (loj instanceof XATMultiSourceOperator) {
            // there is a multi source operator in the middle.
            // cancel the moving.
            // note: because if we want to move, we have to
            //       make sure all the branches got LOJ copied down.
            //       So far, we don't need that.
            // 06/27/2003.
            return;
        }
        loj = loj.getNode().getParent().getOperator();
    }

    if (loj == null)
        return;

    //move loj under the node.
    XATNode parent = loj.getNode().getParent();
    XATNode rightChild = loj.getNode().getChild(1); //right child for loj

    rightChild.setParent(parent);
    for (int i = 0; i < parent.getChildren().length; i++)
        if (parent.getChild(i) == loj.getNode()) {
            parent.setChild(i, rightChild);
            break;
        }

    XATNode taggerChild = xOp.getNode().getChild(0); //only one child of tagger.

    taggerChild.setParent(loj.getNode());
    loj.getNode().setChild(1, taggerChild); //set as new right child of loj

    loj.getNode().setParent(xOp.getNode());
    xOp.getNode().setChild(0, loj.getNode());

}
/**
 * Access method of field context of outter for bindings.
 *
 * @param input array of String
 * @since 1.0
 */
private void setForContext(String[] input) {
    contextForBindings = input;
}
/**
 * Dummy function for expose operator
 * 
 * @return java.lang.Object
 * @param expNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.ExposeImp
 * @since 1.0
 */
public Object visit(Expose expNode) {

    if (isDebug())
        System.out.println("Expose added to decorrelation tree\n");

    return null;
}
/**
 * When a FOR node is encountered, save its child as a temporary
 * node.  This will be the root of a mini-tree.  Then, visit the FOR
 * node's subtree, and look for subsequent values of FOR nodes.
 * If more than one FOR node exits, a Cartesian product must be created.
 * If only one FOR node exists, the mini-tree and the FOR node's subtree
 * can be directly combined.
 * 
 * @return java.lang.Object
 * @param forOp FOR operator
 * @since 1.0
 */
public Object visit(FOR forOp) {
    String forBinding = forOp.getColumnNames()[0];

    // keep the old for context.
    String[] oldForContext = getForContext();

    // keep it global for the later on to generate the groupby operator.
    if (getForContext() == null) {
        setForContext(forOp.getColumnNames());
    } else {
        addForContext(forOp.getColumnNames());
    }

    /** decorrelate the inner query. */
    //go through the subTree
    XATTree subtree = forOp.getXtreFunctions();
    // use the decorrelation object to decorrelate the subtree of the FOR operator.
    // it uses a different constructor, which will inheritant the ForContext over into the
    // new QueryDecorrelation Module.
    QueryDecorrelation subTreeDecor =
        new QueryDecorrelation(this.getForContext(), subtree);

    /** decorrelated the outerQuery. */

    // make the outer query as a subtree.
    XATTree outerTree = new XATTreeImp();
    outerTree.setRoot(forOp.getNode().getChildren()[0]);

    QueryDecorrelation outerTreeDecor =
        new QueryDecorrelation(oldForContext, outerTree);

    // modify the subtree's decorrelated tree based in 4 cases.
    // basically creating cartesion products for additional sources in the sourceVector.
    // we moving this work into complexDecorrelation().
    // subTreeDecor.updateDecorrSubTree(this, outerTreeDecor, forOp);

    // link the updated subTreeDecor back to the current decorrelated tree.
    // depends on # of sources, it might create an outer join, or just linked linear.
    this.complexDecorrelation(outerTreeDecor, forOp, subTreeDecor);

    // fixing the bug of multiple leaves. restore old for context.
    setForContext(oldForContext);
    return null;
}
/**
 * Called when an aggregate function exists in the tree.
 * A GroupBy node replaces the Function node, and an
 * OuterJoin node replaces the FOR node.  A Distinct node
 * is created underneath the OuterJoin node.  The two children
 * of the OuterJoin are the Distinct node and the root of the
 * subtree.
 * 
 * @return java.lang.Object
 * @param function edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp
 * @since 1.0
 */
public Object visit(Function function) {

    if (this.getForContext() != null) {
        if (this.getForContext().length != 0) {
            // for the function there will be a groupby operator created with the FOR context,
            // and then put the function as the subtree of it.

            // in terms of the decorrelated tree, the function node will be deleted from its parent.
            // and then the groupby will take its position.

            XATOperator gbOp = createGBOp(function);
            gbOp.setNode(function.getNode());

            // replace the aggregate operator with the gbNode.
            function.getNode().setOperator(gbOp);
        }
    }

    if (isDebug())
        System.out.println("Function added to decorrelation tree");
    return null;
}
/**
 * Dummy function for Rename operator
 * 
 * @return java.lang.Object
 * @param renNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.RenameImp
 * @since 1.0
 */
public Object visit(NameColumn renNode) {

    if (isDebug())
        System.out.println("NameColumn added to decorrelation tree\n");

    if ("return".equals(renNode.getBlockDesc())) {
        if (isDebug())
            System.out.println("LOJ pushed down to Block(NameColumn)");
        //NOTE: To solve empty collection bug,
        moveLOJ(renNode);
    }

    return null;
}
/**
 * Inserts a source node into the newly decorrelated tree.  If there are
 * two source nodes, then a Cartesian Product node is created.
 * 
 * @return java.lang.Object
 * @param sourceNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.SourceImp
 * @since 1.0
 */
public Object visit(Source sourceOp) {

    if (isDebug())
        System.out.println("Source added to decorrelation tree");

    return null;
}
/**
 * for the groupby there will be a groupby operator created with the FOR context,
 * and then put the original groupby as the subtree of it.
 * in terms of the decorrelated tree, the old groupby node will be deleted from its parent.
 * and then the new groupby will take its position.
 * 
 * @return java.lang.Object
 * @param groupby edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupByImp
 * @since 1.0
 */
public Object visit(GroupBy groupby) {

    if (this.getForContext() != null) {
        if (this.getForContext().length != 0) {

            XATOperator gbOp = createGBOp(groupby);
            gbOp.setNode(groupby.getNode());

            // replace the aggregate operator with the gbNode.
            groupby.getNode().setOperator(gbOp);
        }
    }

    if (isDebug())
        System.out.println("Aggregate node added to decorrelation tree");

    return null;
}
/**
 * Dummy function for intersect operator
 * 
 * @return java.lang.Object
 * @param intersect edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.IntersectImp
 * @since 1.0
 */
public Object visit(Intersection intersect) {

	if(isDebug())System.out.println("Intersect added to decorrelation tree");
	return null;
}
/**
 * Dummy function for orderby operator
 * 
 * @return java.lang.Object
 * @param orderBy edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderByImp
 * @since 1.0
 */
public Object visit(OrderBy orderBy) {

    if (isDebug())
        System.out.println("Orderby added to decorrelation tree");

    return null;
}
/**
 * Dummy function for select operator
 * 
 * @return java.lang.Object
 * @param selectNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.SelectImp
 * @since 1.0
 */
public Object visit(Select selectNode) {

	if(isDebug())System.out.println("Select added to decorrelation tree");
	
	return null;
}
/**
 * Dummy function for union operator
 * 
 * @return java.lang.Object
 * @param union edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.UnionImp
 * @since 1.0
 */
public Object visit(Union union) {

    if (isDebug())
        System.out.println("Union added to decorrelation tree");
    return null;
}
/**
 * Dummy function for delete operator
 * 
 * @return java.lang.Object
 * @param delNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.DeleteImp
 * @since 1.0
 */
public Object visit(Delete delNode) {

    if (isDebug())
        System.out.println("Delete added to decorrelation tree\n");

    return null;
}
/**
 * Dummy function for insert operator
 * 
 * @return java.lang.Object
 * @param insNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.InsertImp
 * @since 1.0
 */
public Object visit(Insert insNode) {

    if (isDebug())
        System.out.println("Insert added to decorrelation tree\n");

    return null;
}
/**
 * Dummy function for replace operator
 * 
 * @return java.lang.Object
 * @param repNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.ReplaceImp
 * @since 1.0
 */
public Object visit(Replace repNode) {

    if (isDebug())
        System.out.println("Replace added to decorrelation tree\n");

    return null;
}
/**
 * for the aggregate function there will be a groupby operator
 * created with the FOR context, and then put the aggregate 
 * function as the subtree of it. In terms of the decorrelated tree, 
 * the aggregate node will be deleted from its parent. and then the 
 * groupby will take its position.
 *
 * @return java.lang.Object
 * @param aggregate edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.AggregateImp
 * @since 1.0
 */
public Object visit(Aggregate aggregate) {

    if (this.getForContext() != null) {
        if (this.getForContext().length != 0) {
            XATOperator gbOp = createGBOp(aggregate);
            gbOp.setNode(aggregate.getNode());

            // replace the aggregate operator with the gbNode.
            aggregate.getNode().setOperator(gbOp);
        }
    }

    if (isDebug())
        System.out.println("Aggregate node added to decorrelation tree");

    return null;
}
/**
 * Dummy function for navigate collection operator
 * 
 * @return java.lang.Object
 * @param navNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollectionImp
 * @since 1.0
 */
public Object visit(NavCollection navNode) {

    if (isDebug())
        System.out.println("Navigate Collection added to decorrelation tree\n");

    return null;
}
/**
 * Dummy function for navigate unnest operator
 * 
 * @return java.lang.Object
 * @param navNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnestImp
 * @since 1.0
 */
public Object visit(NavUnnest navNode) {

    if (isDebug())
        System.out.println("Navigate Unnest added to decorrelation tree\n");

    return null;
}
/**
 * Dummy function for tagger operator
 * 
 * @return java.lang.Object
 * @param taggerNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.TaggerImp
 * @since 1.0
 */
public Object visit(Tagger taggerNode) {

    if (isDebug())
        System.out.println("Tagger added to decorrelation tree");

    //NOTE: To solve empty collection bug,
    //We decided to move Tagger Node (and other constructed (like sortby) node)
    //upper above the LOJ.
    //AND:  If there is a GB between LOJ and Tagger. this kind of move is canceled.
    //      Because there only two place to generate empty collection now: GB+agg and LOJ
    //		We do not need to move it then.
    moveLOJ(taggerNode);

    return null;
}
}
